#// 
#//------------------------------------------------------------------------------
#//   Copyright 2007-2011 Mentor Graphics Corporation
#//   Copyright 2007-2011 Cadence Design Systems, Inc.
#//   Copyright 2010-2011 Synopsys, Inc.
#//   Copyright 2013      NVIDIA Corporation
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//------------------------------------------------------------------------------
#
#`ifndef UVM_DAP_SVH
# `define UVM_DAP_SVH
#
#// Set/Get DAPS
# `include "dap/uvm_set_get_dap_base.svh"
# `include "dap/uvm_simple_lock_dap.svh"
# `include "dap/uvm_get_to_lock_dap.svh"
# `include "dap/uvm_set_before_get_dap.svh"
#
#`endif // UVM_DAP_SVH
#
