
from .uvm_tlm2_defines import *
from .uvm_tlm2_exports import *
from .uvm_tlm2_generic_payload import *
from .uvm_tlm2_ifs import *
from .uvm_tlm2_imps import *
from .uvm_tlm2_ports import *
from .uvm_tlm2_sockets import *
from .uvm_tlm2_sockets_base import *
from .uvm_tlm_time import *
